"""
    Main GUI.
"""
import copy
from functools import partial

import requests
import xbmcgui
import xbmcplugin

from resources.lib.api.api import API
from resources.lib.api.open_subtitles import OpenSubtitles
from resources.lib.auth.auth_provider import AuthProvider
from resources.lib.const import SETTINGS, ROUTE, LANG, ACTION, \
    HTTP_METHOD, COUNT, MEDIA_TYPE, MEDIA_SERVICE, languages, RENDER_TYPE, \
    CONTEXT_MENU, country_lang, explicit_genres, genre_lang, STRINGS, GENERAL, language_lang, sort_methods
from resources.lib.gui import InfoDialog, InfoDialogType
from resources.lib.gui.renderers.context_menu_renderer import ContextMenuRenderer
from resources.lib.gui.renderers.dialog_renderer import DialogRenderer
from resources.lib.gui.renderers.directory_renderer import DirectoryRenderer, Directory
from resources.lib.gui.renderers.icon_renderer import IconRenderer
from resources.lib.gui.renderers.media_info_renderer import MediaInfoRenderer
from resources.lib.gui.renderers.media_list_renderer import MediaListRenderer
from resources.lib.gui.renderers.trakt_renderer import TraktRenderer
from resources.lib.kodilogging import logger
from resources.lib.routing.command import Command
from resources.lib.routing.router import router
from resources.lib.storage.settings import settings
from resources.lib.storage.sqlite import DB
from resources.lib.stream_picker import StreamPicker
from resources.lib.utils.csfd_tips import get_csfd_tips
from resources.lib.utils.kodiutils import get_string, kodi_json_request, key_combiner, \
    get_current_alphabet_index, is_youtube_url
from resources.lib.api.trakt_api import trakt
from resources.lib.utils.url import Url
from resources.lib.utils.youtube import get_yt_video_url


class StreamCinema:
    def __init__(self, provider, api):
        self.provider = provider
        self._api = api
        self.auth = AuthProvider(provider)

        self.directory_renderer = DirectoryRenderer()
        self.media_renderer = MediaListRenderer()

        self.route_map = {
            ROUTE.ROOT: self.main_menu,
            ROUTE.SEARCH: self.search,
            ROUTE.GET_MEDIA: self.get_media,
            ROUTE.MOVIES: self.movies,
            ROUTE.TV_SHOWS: self.tv_shows,
            ROUTE.COMMAND: self.command,
            ROUTE.PROCESS_MEDIA_ITEM: self.process_media_item,
            ROUTE.COUNT_MENU: self.count_menu,
            ROUTE.CSFD_TIPS: self.csfd_tips,
            ROUTE.WATCHED: self.watched,
            ROUTE.CLEAR_PATH: self.clear_path,
            ROUTE.DOWNLOAD_QUEUE: self.download_queue,
            ROUTE.PLAY_TRAILER: self.play_trailer
        }

        if trakt.is_authenticated():
            self.trakt_renderer = TraktRenderer(self)
            self.route_map.update({
                ROUTE.TRAKT_HISTORY_LIST: self.trakt_renderer.history_lists,
                ROUTE.TRAKT_FRIENDS_LIST: self.trakt_renderer.friends_list,
                ROUTE.TRAKT_LIST: self.trakt_renderer.list,
                ROUTE.TRAKT_LISTS: self.trakt_renderer.lists,
                ROUTE.TRAKT_SPECIAL_LIST: self.trakt_renderer.special_list,
                ROUTE.TRAKT_HISTORY: self.trakt_renderer.history
            })

        self.dir_type_render = {
            RENDER_TYPE.DEFAULT: self.get_media_default,
            RENDER_TYPE.A_Z: self.get_media_az,
            RENDER_TYPE.SEARCH + '?': self.get_media_search
        }

        self.context_menu_map = {
            ROUTE.ROOT: [],
            ROUTE.WATCHED: [
                CONTEXT_MENU.DELETE_HISTORY,
                CONTEXT_MENU.DOWNLOAD,
                CONTEXT_MENU.CHOOSE_STREAM
            ],
            ROUTE.COUNT_MENU: [
                CONTEXT_MENU.PIN
            ],
            ROUTE.GET_MEDIA: [
                CONTEXT_MENU.DOWNLOAD,
                CONTEXT_MENU.ADD_TO_LIBRARY,
                CONTEXT_MENU.TRAKT,
                CONTEXT_MENU.CHOOSE_STREAM
            ],
            ROUTE.DOWNLOAD_QUEUE: [
                CONTEXT_MENU.PAUSE_DOWNLOAD,
                CONTEXT_MENU.DELETE_DOWNLOAD,
                CONTEXT_MENU.START_DOWNLOAD
            ],
            ROUTE.TRAKT_LIST: [
                CONTEXT_MENU.TRAKT
            ]
        }

        self.context_menu = {
            CONTEXT_MENU.DELETE_HISTORY: ContextMenuRenderer.get_context_menu_history,
            CONTEXT_MENU.PIN: ContextMenuRenderer.pin,
            CONTEXT_MENU.DOWNLOAD: ContextMenuRenderer.get_context_menu_download,
            CONTEXT_MENU.DELETE_DOWNLOAD: ContextMenuRenderer.delete_download_item,
            CONTEXT_MENU.START_DOWNLOAD: ContextMenuRenderer.start_download_item,
            CONTEXT_MENU.PAUSE_DOWNLOAD: ContextMenuRenderer.pause_download_item,
            CONTEXT_MENU.ADD_TO_LIBRARY: ContextMenuRenderer.add_to_library_item,
            CONTEXT_MENU.CHOOSE_STREAM: ContextMenuRenderer.choose_stream,
            CONTEXT_MENU.TRAKT: ContextMenuRenderer.trakt,
        }

        self.count_menu_map = {
            COUNT.COUNTRIES: {
                'title': DirectoryRenderer.TITLE.count,
                'icon': IconRenderer.country,
                'translate': country_lang,
            },
            COUNT.GENRES: {
                'title': DirectoryRenderer.TITLE.count,
                'icon': IconRenderer.default,
                'translate': genre_lang,
                'filter': lambda x: not settings.as_bool(
                    SETTINGS.EXPLICIT_CONTENT) and x not in explicit_genres or settings.as_bool(
                    SETTINGS.EXPLICIT_CONTENT)
            },
            COUNT.TITLES: {
                'title': DirectoryRenderer.TITLE.count,
                'icon': IconRenderer.a_z,
                'url': DirectoryRenderer.URL.az,
                'process_menu': DirectoryRenderer.a_to_z_menu
            },
            COUNT.YEARS: {
                'title': DirectoryRenderer.TITLE.count,
                'sort_reverse': True,
                'sort': lambda x: int(x.get('translated_key'))
            },
            COUNT.LANGUAGES: {
                'title': DirectoryRenderer.TITLE.count,
                'translate': language_lang,
            }
        }

        self.command_center = Command(self)
        self.stream_picker = StreamPicker(self)

    def get_context_menu(self, route):
        context_menu = self.context_menu_map.get(route, [])
        return [partial(self.context_menu[ctx_m], route) for ctx_m in context_menu]

    def dispatch(self, handle, url, query):
        StreamCinema.add_sort_methods(handle)
        self.route_map[url](handle, route=url, **dict(Url.parse_qsl(query.lstrip('?'))))

    @staticmethod
    def add_sort_methods(handle):
        for method in sort_methods:
            xbmcplugin.addSortMethod(handle, method)

    def command(self, handle, command, route, **kwargs):
        xbmcplugin.endOfDirectory(handle)
        self.command_center(command)(**kwargs)

    def main_menu(self, handle, route):
        self.directory_renderer(handle, Directory.main_menu(), route)

    def movies(self, handle, route):
        self.directory_renderer(handle, Directory.movies(), route)

    def tv_shows(self, handle, route):
        self.directory_renderer(handle, Directory.tv_shows(), route)

    def count_menu(self, handle, count_type, filter_type, render_type, route, filter_value=STRINGS.EMPTY,
                   media_type=None, page=0):
        count_map = self.count_menu_map.get(count_type, {})

        response = self.api_request(HTTP_METHOD.GET,
                                    count_map.get('api_url', self._api.FILTER.count)(media_type, count_type,
                                                                                     filter_value, page))
        icon_renderer = count_map.get('icon', IconRenderer.default)
        title_renderer = count_map.get('title', DirectoryRenderer.TITLE.default)
        url_builder = partial(count_map.get('url', DirectoryRenderer.URL.default), media_type, filter_type, render_type,
                              count_type)
        sort_key = key_combiner(count_map.get('sort', lambda x: get_current_alphabet_index(x.get('translated_key'))))

        data = response.get('data', [])
        pinned_items = [row[0] for row in DB.PINNED.get(route + count_type)] if route else []
        data_keys = [item.get('key') for item in data]
        for item in pinned_items:
            if item not in data_keys:
                data.append({'key': item, 'doc_count': None})

        count_lang = count_map.get('translate', {})
        processed_data = []
        for item in data:
            key = item.get('key')
            lang_id = count_lang.get(key)
            title = get_string(lang_id) if lang_id else key.encode('utf-8')
            item.update({'title': title, 'translated_key': title})
            item.update({'title': title_renderer(item)})
            if count_map.get('filter', lambda *args: True)(key):
                processed_data.append(item)

        processed_data.sort(key=sort_key, reverse=count_map.get('sort_reverse', False))
        context_menu = self.get_context_menu(ROUTE.COUNT_MENU)
        count_menu = Directory.count_menu(processed_data, icon_renderer, context_menu, url_builder, count_type)
        count_menu = count_map.get('process_menu', lambda *args: count_menu)(media_type, count_menu, filter_value,
                                                                             response)
        if len(data) - len(pinned_items) == GENERAL.DIR_PAGE_LIMIT:
            next_page = Directory.count_menu_next_page(media_type, count_type, filter_type, render_type, filter_value,
                                                       int(page) + GENERAL.DIR_PAGE_LIMIT)
            MediaListRenderer.add_navigation(count_menu, True)
            count_menu.append(next_page)
        self.directory_renderer(handle, count_menu, route + count_type)

    def search(self, handle, route, media_type=MEDIA_TYPE.ALL):
        self.directory_renderer.search(handle, media_type)

    def get_media(self, handle, route, url, render_type, media_type, **kwargs):
        self.dir_type_render[render_type](handle, route, url, render_type, media_type, **kwargs)

    def get_media_default(self, handle, route, url, render_type, media_type):
        response = self.api_request(HTTP_METHOD.GET, url)
        self.render_media(handle, route, response, render_type=render_type, media_type=media_type)

    def get_media_search(self, handle, route, url, render_type, dir_type):
        response = self.api_request(HTTP_METHOD.GET, url)
        self.render_media(handle, route, response, MediaInfoRenderer.TITLE.mixed, render_type, dir_type)

    def get_media_az(self, handle, route, url, render_type, media_type, letters):
        media_list = self.api_request(HTTP_METHOD.GET, url)
        context_menu = self.get_context_menu(route)
        media_list_gui = self.media_renderer.build_media_list(media_list, RENDER_TYPE.DEFAULT, context_menu,
                                                              MediaInfoRenderer.TITLE.a_z, letters)
        media_list_gui.sort(key=lambda x: x.sort_title)
        self._render_call(handle, media_list, media_list_gui, render_type, media_type)

    def render_media(self, handle, route, media_list, title_renderer=MediaInfoRenderer.TITLE.default,
                     render_type=RENDER_TYPE.DEFAULT, media_type=MEDIA_TYPE.ALL):
        context_menu = self.get_context_menu(route)

        media_list_gui = self.media_renderer.build_media_list(media_list, render_type, context_menu,
                                                              title_renderer)
        self._render_call(handle, media_list, media_list_gui, render_type, media_type)

    def _render_call(self, handle, media_list, media_list_gui, render_type=RENDER_TYPE.DEFAULT,
                     media_type=MEDIA_TYPE.ALL):
        self.media_renderer(handle, media_list_gui, media_list.get('pagination'), render_type, media_type)

    def csfd_tips(self, handle, route):
        url = API.FILTER.service(MEDIA_TYPE.MOVIE, MEDIA_SERVICE.CSFD, get_csfd_tips())
        response = self.api_request(HTTP_METHOD.GET, url)
        self.render_media(handle, route, response)

    def process_media_item(self, handle, route, media_id, url, root_parent_id, action=ACTION.PLAY, force='0'):
        selected_stream = self.stream_picker.get_and_select_stream(url, force)
        if selected_stream:
            self.stream_picker.stream_guard(handle, selected_stream, media_id, root_parent_id, action)
        else:
            router.set_resolved_url(handle)

    def api_request(self, *args, **kwargs):
        return self.api_response_handler(self._api.request(*args, **kwargs))

    def watched(self, handle, route):
        results = DB.WATCH_HISTORY.get_all()
        ids = [row[0] for row in results]
        logger.debug('Getting history list')
        api_res = self.api_request(HTTP_METHOD.GET, self._api.FILTER.ids(ids))
        order = {v: i for i, v in enumerate(ids)}
        api_res['data'] = sorted(api_res['data'], key=lambda x: order[x['_id']])
        self.render_media(handle, route, api_res)

    def download_queue(self, handle, route):
        api_res = self.get_downloading_items()
        self.render_media(handle, route, media_list=api_res, title_renderer=MediaInfoRenderer.TITLE.download_queue,
                          render_type=RENDER_TYPE.DOWNLOAD)

    def get_downloading_items(self):
        results = DB.DOWNLOAD.get_all()
        ids = list(set([row[9] for row in results]))
        api_result = self.api_request(HTTP_METHOD.GET, self._api.FILTER.ids(ids))
        temp = {}
        for item in api_result['data']:
            temp[item['_id']] = item
        data = []
        for row in results:
            media_id = row[9]
            if media_id in temp:
                api = copy.deepcopy(temp[media_id])
                api['dl_id'] = row[0]
                api['_source']['download_perc'] = row[5]
                api['_source']['download_speed'] = row[6]
                data.append(api)
        api_result.update({'data': data})
        return api_result

    @staticmethod
    def api_response_handler(response):
        try:
            response.raise_for_status()
            return response.json()
        except requests.exceptions.HTTPError as e:
            InfoDialog(get_string(LANG.HTTP_ERROR), icon=InfoDialogType.ERROR).notify()
            logger.error(e)
            return None

    @staticmethod
    def has_preferred_subtitles(stream):
        subs = stream.get('subtitles')
        for sub in subs:
            if sub.get('language') == settings.get_select(SETTINGS.SUBTITLES_PROVIDER_PREFERRED_LANGUAGE):
                logger.debug('Found desired subtitles in the stream')
                return True
        return False

    @staticmethod
    def get_subtitles(media):
        if not media.get('services', {}).get('imdb'):
            return

        try:
            subs = OpenSubtitles.search(media)
        except:
            return

        if subs:
            subs_path = OpenSubtitles.download(subs)
            return subs_path

    @staticmethod
    def send_service_message(service_name, service_event, **kwargs):
        logger.debug('Sending service message {}: {}'.format(service_name, service_event))
        kodi_json_request(service_name, service_event, kwargs)

    @staticmethod
    def set_language(title, language_type):
        languages_values = languages.values()
        selected = DialogRenderer.select(title, languages_values)
        if selected is None:
            return
        selected_language = languages_values[selected]
        selected_language_code = [k for k, v in languages.items() if v == selected_language][0]
        settings.set_select(language_type, selected_language_code)

    def intro(self):
        # WELCOME
        DialogRenderer.ok(get_string(LANG.INTRO_TITLE_1), get_string(LANG.INTRO_TEXT_1))
        # LANGUAGE
        DialogRenderer.ok(get_string(LANG.INTRO_TITLE_2), get_string(LANG.INTRO_TEXT_2))
        StreamCinema.set_language(get_string(LANG.CHOOSE_LANGUAGE), SETTINGS.PREFERRED_LANGUAGE)
        StreamCinema.set_language(get_string(LANG.CHOOSE_FALLBACK_LANGUAGE), SETTINGS.FALLBACK_LANGUAGE)
        # SUBTITLES
        yes = DialogRenderer.yesno(get_string(LANG.INTRO_TITLE_3), get_string(LANG.INTRO_TEXT_3))
        if yes:
            DialogRenderer.ok(get_string(LANG.INTRO_TITLE_3), get_string(LANG.INTRO_TEXT_4))
            DialogRenderer.set_subtitles_credentials()
            StreamCinema.set_language(get_string(LANG.INTRO_TITLE_4), SETTINGS.SUBTITLES_PROVIDER_PREFERRED_LANGUAGE)
            StreamCinema.set_language(get_string(LANG.INTRO_TITLE_5), SETTINGS.SUBTITLES_PROVIDER_FALLBACK_LANGUAGE)
        # PROVIDER
        DialogRenderer.ok(get_string(LANG.INTRO_TITLE_6), get_string(LANG.INTRO_TEXT_7))
        self.auth.set_provider_credentials()
        # SEARCH INFO
        DialogRenderer.ok(get_string(LANG.INTRO_TITLE_7), get_string(LANG.INTRO_TEXT_8))
        # ENDING
        DialogRenderer.ok(get_string(LANG.INTRO_TITLE_8), get_string(LANG.INTRO_TEXT_9))

    @staticmethod
    def clear_path(handle, route):
        logger.debug('Clearing path')
        router.replace_route(handle, ROUTE.ROOT)

    @staticmethod
    def play_trailer(handle, route, video_url):
        logger.debug('Getting trailer URL %s' % video_url)
        if is_youtube_url(video_url):
            video_url = get_yt_video_url(video_url)
        return router.set_resolved_url(handle, xbmcgui.ListItem(path=video_url))
